SystemOrganization addCategory: #'DoItFirst-System-Support'!

Object subclass: #DoItFirst
	instanceVariableNames: 'actions allowInitialization'
	classVariableNames: 'Current'
	poolDictionaries: ''
	category: 'DoItFirst-System-Support'!

!DoItFirst commentStamp: 'dtl 6/14/2020 08:56' prior: 0!
Be the first thing in the system startup list, and do things that should be done prior to any additional image initialization. If the first image argument is a recognized option, evaluate it. Image arguments are typically preceded by a '--' token on the command line.

Possible command options:
   -filein {some file names}
   -doit {some strings}
   -evaluate {a single string & write result to stdout & quit

!

!DoItFirst class methodsFor: 'instance creation' stamp: 'dtl 6/15/2020 17:34'!
current
	^ Current! !

!DoItFirst class methodsFor: 'reevaluate options' stamp: 'dtl 6/15/2020 20:20'!
evaluate: optionSymbol
	"If optionSymbol was registered during command option parsing, reevaluate its
	option block now."

	^self current evaluate: optionSymbol! !

!DoItFirst class methodsFor: 'class initialization' stamp: 'dtl 6/11/2020 18:42'!
initialize

	"DoItFirst initialize"

	Smalltalk addToStartUpList: self before: SmallInteger.! !

!DoItFirst class methodsFor: 'reevaluate options' stamp: 'dtl 6/15/2020 20:20'!
reevaluateCwd
	"If a -cwd option was specified on the command line, reevaluate it now. Assuming
	that DoItFirst class>>startUp: is called early in the startup sequence, the default
	directory might later be set by FileDirectory class>>startUp:. This method provides
	a hook for FileDirectory class>>startUp: to reevaluate the command line option to
	ensure that the default directory is ultimately set as specified by the -cwd image
	command line option."

	"DoItFirst reevaluateCwd"

	^self evaluate: #cwd.! !

!DoItFirst class methodsFor: 'reevaluate options' stamp: 'dtl 8/6/2020 22:08'!
reevaluateDebug
	"The -debug option cannot be evaluated at DoInNow startUp time, but may be called
	later in the startUp processing. If -debug was not specified as a command option this
	method does nothing.

	May be called from Delay>>startup to invoke a debugger at the earliest possible time."

	"DoItFirst reevaluateDebug"

	^self evaluate: #debug.! !

!DoItFirst class methodsFor: 'system startup' stamp: 'dtl 6/15/2020 18:07'!
startUp: resuming
	resuming ifTrue: [ Current := self new. Current evaluate ]
! !

!DoItFirst methodsFor: 'private' stamp: 'dtl 6/15/2020 17:59'!
add: valuable to: actionList at: key
	"Add valuable to the end of action list, and register it at key so that it may
	be reevaluated at a later time if necessary."
	actionList addLast: valuable.
	actions at: key put: valuable.! !

!DoItFirst methodsFor: 'private' stamp: 'dtl 6/15/2020 17:59'!
addFirst: valuable to: actionList at: key
	"Add valuable to the beginning of action list, and register it at key so that it may
	be reevaluated at a later time if necessary."
	actionList addFirst: valuable.
	actions at: key put: valuable.! !

!DoItFirst methodsFor: 'private' stamp: 'dtl 8/6/2020 20:17'!
addWithoutEvaluation: valuable at: key
	"Register action at key so that it may be reevaluated at a later time. Do not
	evaluate in the startUp of DoItNow. Used when the action cannot yet be
	evaluated because it requires startUp processing later in the startup list."
	actions at: key put: valuable.! !

!DoItFirst methodsFor: 'actions' stamp: 'dtl 6/14/2020 12:36'!
cwd: path
	"Evaluate arg and print the result on stdout, or error message on stderr.
	Exit immediately without saving the image."
	(FileDirectory on: path) exists
		ifTrue: [ FileDirectory setDefaultDirectory: path ]
		ifFalse: [ FileStream stderr nextPutAll: path, ': directory does not exist'; lf; flush.
				Smalltalk quitPrimitive ]
! !

!DoItFirst methodsFor: 'actions' stamp: 'dtl 8/6/2020 20:53'!
debug
	"halt and enter a debugger"
	self halt: self class name, ' inserted break in StartUpList processing'.! !

!DoItFirst methodsFor: 'actions' stamp: 'dtl 6/14/2020 12:36'!
doIt: arguments
	"Evaluate arguments and print the result on stdout, or error message on stderr.
	Exit the image after any error."
	arguments do: [ :arg |
		[FileStream stdout nextPutAll: (Compiler evaluate: arg) asString; lf; flush]
			on: Error
			do: [ :ex | FileStream stderr nextPutAll: ex asString; lf; flush.
				Smalltalk quitPrimitive ]]! !

!DoItFirst methodsFor: 'evaluating' stamp: 'dtl 8/6/2020 19:13'!
evaluate
	| actionQueue |
	actionQueue := self parse readStream.
	actionQueue isEmpty
		ifFalse: [ allowInitialization
			ifTrue: [
				FileStream startUp: true. "initialize stdout and stderr"
				FileDirectory startUp "set default directory" ].
			[ actionQueue atEnd ] whileFalse: [ actionQueue next value ] ].
! !

!DoItFirst methodsFor: 'evaluating' stamp: 'dtl 6/15/2020 19:49'!
evaluate: actionKey
	"If actionKey is registered, then evaluate its action. This is intended to allow
	a previously evaluated option to be reevaluated at a later point in the system
	startup list if necessary."
	(actions at: actionKey ifAbsent: []) ifNotNil: [:action | action value]! !

!DoItFirst methodsFor: 'actions' stamp: 'dtl 7/17/2020 21:54'!
evaluateArg: arg
	"Evaluate arg and print the result on stdout, or error message on stderr.
	Exit immediately without saving the image."
	[FileStream stdout nextPutAll: (Compiler evaluate: arg) asString; lf; flush]
		on: Error
		do: [ :ex | FileStream stderr nextPutAll: ex asString; lf; flush ].
	Smalltalk quitPrimitive! !

!DoItFirst methodsFor: 'actions' stamp: 'dtl 6/14/2020 12:36'!
fileIn: fileNames
	"File in each named file. On error, print a message to stderr and exit the image."
	fileNames do: [ :arg |
		[ | fs |
		fs := FileStream oldFileNamed: arg.
		FileStream stdout nextPutAll: 'fiie in ', fs asString; lf; flush.
		fs fileIn ]
			on: Error
			do: [ :ex | FileStream stderr nextPutAll: ex asString; lf; flush.
				Smalltalk quitPrimitive ]]! !

!DoItFirst methodsFor: 'actions' stamp: 'dtl 8/6/2020 22:55'!
help
	FileStream stdout nextPutAll: self class name, ' image arguments:'; lf.
	{	'-doit argumentlist "evaluate each argument as a doIt expression"' .
		'-evaluate arg "evaluate arg, print result then exit"' .
		'-filein filelist "file in each file named in fileList"' .
		'-cwd path "set FileDirectory defaultDirectory to path prior to evaluating other options"' .
		'-debug "enter a debugger as soon as possible in the startUp processing"'.
		'-noinit "mimimal initialization, suppress file and directory startUp processing when evaluating options"'.
		'-help "print this message"'
	} do: [ :e | FileStream stdout tab; nextPutAll: e; lf ].
	FileStream stdout flush.
	Smalltalk quitPrimitive.

! !

!DoItFirst methodsFor: 'initialize-release' stamp: 'dtl 8/6/2020 19:09'!
initialize
	actions := Dictionary new.
	allowInitialization := true.
! !

!DoItFirst methodsFor: 'initialize-release' stamp: 'dtl 6/14/2020 12:03'!
isArg: token
	^ token isEmpty not and: [ token beginsWith: '-' ].! !

!DoItFirst methodsFor: 'evaluating' stamp: 'dtl 6/14/2020 12:11'!
nextTokensFrom:  argumentStream
	"Next available tokens up to the next parseable argument, for commands
	that expect an argument list of names or doIt expressions."

	| list |
	list := OrderedCollection new.
	[ argumentStream atEnd or: [ self isArg: argumentStream peek ]]
		whileFalse: [ list add: argumentStream next ].
	^ list! !

!DoItFirst methodsFor: 'actions' stamp: 'dtl 8/6/2020 19:20'!
noInit
	"Do not allow other actions to call startUp processing for file and directory initialiation.
	Standard input and output stream will not be available. This allows the command line
	to specify a minimal startUp initialization, which may be useful in the case of a badly
	broken image."

	allowInitialization := false
! !

!DoItFirst methodsFor: 'evaluating' stamp: 'dtl 8/6/2020 21:45'!
parse
	| args actions |
	args := Smalltalk arguments readStream.
	actions := OrderedCollection new.
	[ args atEnd ] whileFalse: [
		args next ifNotNil: [ :next |
			(next caseOf: {
				[ '-noinit' ] -> [ self noInit] .
				[ '--noinit' ] -> [ self noInit] .
				[ '-help' ] -> [ self addFirst: [ self help ] to: actions at: #help] .
				[ '--help' ] -> [ self addFirst: [ self help ] to: actions at: #help] .
				[ '-debug' ] -> [ self addWithoutEvaluation: [ self debug ] at: #debug] .
				[ '--debug' ] -> [ self addWithoutEvaluation: [ self debug ] at: #debug] .
				[ '-doit' ] -> [ | list | list := self nextTokensFrom: args. self add:[ self doIt: list ] to: actions at: #doit] .
				[ '--doit' ] -> [ | list | list := self nextTokensFrom: args. self add:[ self doIt: list ] to: actions at: #doit] .
				[ '-evaluate' ] -> [ | arg | arg := args next.  self add:[ self evaluateArg: arg ] to: actions at: #evaluate] .
				[ '--evaluate' ] -> [ | arg | arg := args next.  self add:[ self evaluateArg: arg ] to: actions at: #evaluate] .
				[ '-filein' ] -> [ | list | list := self nextTokensFrom: args. self add:[ self fileIn: list ] to: actions at: #filein] .
				[ '--filein' ] -> [ | list | list := self nextTokensFrom: args. self add:[ self fileIn: list ] to: actions at: #filein] .
				[ '-cwd' ] -> [ | arg | arg := args next.  self addFirst:[ self cwd: arg ] to: actions at: #cwd] .
				[ '--cwd' ] -> [ | arg | arg := args next.  self addFirst:[ self cwd: arg ] to: actions at: #cwd] .
			} otherwise: [^ actions ] ) ] ].
	^ actions.
! !
DoItFirst initialize!
